/* This file is part of Jumper
   Copyright (C) 2013 Sergey Poznyakoff
  
   Jumper is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.
  
   Jumper is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with Jumper.  If not, see <http://www.gnu.org/licenses/>.
*/
#include "jumper.h"
#include <grecs.h>

#ifndef SYSCONFDIR
# define SYSCONFDIR "/etc"
#endif
#ifndef LOCALSTATEDIR
# define LOCALSTATEDIR "/var"
#endif
#define DEFAULT_CONFFILE SYSCONFDIR "/jumper.conf"
char *conffile = DEFAULT_CONFFILE;

int debug_level;
int foreground;
int preprocess_only;
char *pidfile = LOCALSTATEDIR "/run/jumper.pid";

int facility = LOGFACILITY;
char *tag;

static int opt_debug_level = 0;
static int opt_foreground = 0;
static char *opt_pidfile = NULL;
static int opt_facility = -1;
static int lint_only = 0;

#define NITEMS(a) (sizeof(a)/sizeof((a)[0]))

int
sigv_set_action(int sigc, int *sigv, struct sigaction *sa)
{
	int i;
	
	for (i = 0; i < sigc; i++) {
		if (sigaction(sigv[i], &sa[i], NULL))
			return i+1;
	}
	return 0;
}

int
sigv_set_all(void (*handler)(int), int sigc, int *sigv,
	     struct sigaction *retsa)
{
	int i;
	struct sigaction sa;
	
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = 0;
	
	for (i = 0; i < sigc; i++) {
		sa.sa_handler = handler;
		
		if (sigaction(sigv[i], &sa, retsa ? &retsa[i] : NULL)) {
			if (retsa) {
				int ec = errno;
				sigv_set_action(i, sigv, retsa);
				errno = ec;
			}
			return -1;
		}
	}
	return 0;
}

void
signal_setup(void (*sf) (int))
{
	static int sigv[] = { SIGTERM, SIGQUIT, SIGINT, SIGHUP, SIGALRM,
			      SIGUSR1, SIGUSR1, SIGCHLD };
	sigv_set_all(sf, NITEMS(sigv), sigv, NULL);
}

void
storepid(const char *pidfile)
{
	FILE *fp = fopen(pidfile, "w");
	if (!fp) {
		diag(LOG_ERR, "cannot open pidfile %s for writing: %s",
		     pidfile, strerror(errno));
	} else {
		fprintf(fp, "%lu\n", (unsigned long) getpid());
		fclose(fp);
	}
}

int signo = 0;
int stop = 0;

void
sigmain(int sig)
{
	signo = sig;
	switch (signo) {
	case SIGCHLD:
		progman_cleanup();
		break;
	case SIGALRM:
		break;
	default:
		stop = 1;
	}
}

#include "cmdline.h"

int
main(int argc, char **argv)
{
	int i;
	struct grecs_node *tree;
	
	set_program_name(argv[0]);
	tag = (char*) program_name;
	config_init();

	parse_options(argc, argv, &i);

	argc -= i;
	argv += i;

	switch (argc) {
	default:
		diag(LOG_CRIT, "too many arguments");
		exit(EX_USAGE);
	case 1:
		conffile = argv[0];
		break;
	case 0:
		break;
	}

        if (preprocess_only)
                exit(grecs_preproc_run(conffile, grecs_preprocessor) ?
                     EX_CONFIG : 0);
	
	tree = grecs_parse(conffile);
	if (!tree)
		exit(EX_CONFIG);

	config_finish(tree);

	if (opt_debug_level)
		debug_level += opt_debug_level;
	if (opt_foreground)
		foreground = opt_foreground;
	if (opt_pidfile)
		pidfile = opt_pidfile;
	if (opt_facility != -1)
		facility = opt_facility;
	if (!foreground && facility <= 0)
		facility = LOG_DAEMON;

	interface_init();
	listener_init();

	if (lint_only)
		return 0;

	if (facility > 0) {
		openlog(tag, LOG_PID, facility);
		grecs_log_to_stderr = 0;
	}

	if (!foreground) {
		if (daemon(0, 0) < 0) {
			diag(LOG_CRIT, "daemon: %s", strerror(errno));
			exit(EX_OSERR);
		}
		log_to_stderr = -1;
	}

	diag(LOG_INFO, "%s %s started", program_name, VERSION);
	storepid(pidfile);

	signal_setup(sigmain);
	listener_run_action(event_startup);

	while (!stop)
		listener_run();

	progman_terminate();
	listener_run_action(event_cleanup);
	
	diag(LOG_INFO, "%s %s stopped", program_name, VERSION);

	unlink(pidfile);
	
	return 0;
}
